/**
 ** A Virtual Host Monitor
 **
 ** Copyright 2009 Gary Wuertz gary@issiweb.com
 **
 ** This program is free software: you can redistribute it and/or modify
 ** it under the terms of the GNU General Public License as published by
 ** the Free Software Foundation, either version 3 of the License, or
 ** (at your option) any later version.
 **
 ** This program is distributed in the hope that it will be useful,
 ** but WITHOUT ANY WARRANTY; without even the implied warranty of
 ** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 ** GNU General Public License for more details.
 **
 ** You should have received a copy of the GNU General Public License
 ** along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include <iostream> // for debugging

#include <algorithm>

#include <errno.h>
#include <glob.h>
#include <sys/inotify.h>

#include "monitor.h"

#define FMT_DEFAULT "%h %l %u %t \\\"%r\\\" %>s %b"
#define EVENT_SIZE  (sizeof (struct inotify_event))
#define BUF_LEN     (1024 * (EVENT_SIZE + 16))
#define PATH_LEN    256

using namespace std;
/**
 * Sort functions for the display
 */
bool byAccess(const vhost& a, const vhost& b)
{
  if (a.ldtm == b.ldtm) return a.hits > b.hits;
  return a.ldtm > b.ldtm;
}
bool byHits(const vhost& a, const vhost& b)
{
  if (a.hits == b.hits) return a.ldtm > b.ldtm;
  return a.hits > b.hits;
}
bool byName(const vhost& a, const vhost& b)
{
  string na = a.fields[a.nidx];
  string nb = b.fields[b.nidx];
  if (na == nb) return a.ldtm > b.ldtm;
  return na < nb;
}
/**
 * Return the hosts list
 */
vhost  * Monitor::getHosts(topvhost *app)
{
  int i, hits;
  long meter=0;

  for(i=hits=0;i<this->nhosts;i++) {
    if (this->hosts[i].dirty && !this->hosts[i].setClean(app))
      this->setError("clean:");
    this->hlist[i] = this->hosts[i];
    hits += this->hosts[i].hits;
    meter += this->hosts[i].meter;
    }
  switch (app->order) {
    case topvhost::BY_ACCESS:
      std::sort(this->hlist,this->hlist+this->nhosts,byAccess);
      break;
    case topvhost::BY_HITS:
      std::sort(this->hlist,this->hlist+this->nhosts,byHits);
      break;
    case topvhost::BY_NAME:
      std::sort(this->hlist,this->hlist+this->nhosts,byName);
      break;
    }
  app->hits    = hits;
  app->meter   = meter;
  app->total   = this->nhosts;
  if (app->maxy!=0) {
    int ct = app->maxy - app->hdrsz;
    app->visible = min(ct,this->nhosts);
    }
  else app->visible = this->nhosts;
  return this->hlist;
}
/**
 * Get the watch list - the list of domains and associated log files. The list
 * is constructed from:
 *
 *   1) "glob_src"   = a file system glob (with optional "glog_omit" exclusion)
 *   2) "file_src"   = the top level section of the specified config file
 *   3) "config_src" = the name of a section in the current config file
 *
 * For configuration files the syntax is <domain name>=<log file path>
 */
watchList* Monitor::getWatchList(Config * conf)
{
  watchList*  wl      = new watchList();
  const char* xfile   = conf->get("file_src").c_str();
  const char* pattern = conf->get("glob_src").c_str();
  const char* xsect   = conf->get("config_src").c_str();
  map<string,string>::iterator iter;
  int         len;

  if (strlen(pattern)>0) {            // handle glob as source
    char        path[PATH_LEN];
    glob_t      p;
    char**      w;
    const char* exclude = conf->get("glob_omit").c_str();
    int         i, offset, overhead, skip;
    string      sl;

    path[PATH_LEN-1] = 0;
    skip    = strlen(exclude);
    snprintf(path,PATH_LEN-1,pattern, "*");
    if (glob(path, GLOB_NOSORT, NULL, &p))
      this->setError("glob:");
    else {
      offset = strstr(pattern,"%s") - pattern;
      overhead = strlen(pattern)-2;
      w = p.gl_pathv;
      for(i = 0;i<(int)p.gl_pathc;i++) {
        len = strlen(w[i]);
        if (skip>0 && len>skip && !strcmp(w[i]+len-skip,exclude)) continue;
        len -= overhead;
        strncpy(path,w[i]+offset,len);
        path[len] = 0;
        sl = string(path);
        snprintf(path,PATH_LEN-1,pattern,sl.c_str());
        wl->watch[sl] = path;
        }
      globfree(&p);
      }
    }
  if (strlen(xfile)) {                  // Handle external file as source
    Config *xconf = new Config(xfile);
    map<string,string>* section = xconf->getSection("");
    for(iter=section->begin();iter!=section->end();++iter)
        wl->watch[iter->first] = iter->second;
    }
  if (strlen(xsect)>0) {              // Handle section within config as source
    map<string,string>* section = conf->getSection(xsect);
    for(iter=section->begin();iter!=section->end();++iter)
        wl->watch[iter->first] = iter->second;
    }
  for(iter=wl->watch.begin();iter!=wl->watch.end();++iter) {
    if ((len=iter->first.length())>wl->namelen)
      wl->namelen = len;
    }
  if (wl->namelen>0)
    wl->success = true;
  else wl->problem = "Empty watch list";
  return wl;
}
/**
 * Initialize the monitor
 */
bool  Monitor::init(topvhost *app, struct monitorOptions options)
{
  vhost * inst;
  int   i, nfiles;
  unsigned j,k,nidx=0;
  string s;
  watchList* wl;
  map<string,string>::iterator iter;

  Config* conf = new Config(options.configFile);
  if (*options.configSection)
    conf->load(options.configSection);

  app->hfmt = conf->get("hdr_format");
  app->rfmt = conf->get("row_format");
  app->logfmt = conf->get("log_format");
  app->hdrsz = 2;                         // temporary for now

  vhost::setVector(app);                  // set needs
  this->parseApacheLogFormat(app);        // set gets

  for(j = 0;j<app->needs.size();j++) {    // xfer maps gets to needs
    string ch  = app->needs[j];
    int idx = vhost::OUTPUT_FIELDS+j;
    if (ch == "v")                        // domain does not come from log record
      nidx = idx;
    else for(k=0;k<app->gets.size();k++)
      if (ch == app->gets[k]) {
        app->xfer[k] = idx;
        break;
        }
    }
  if (app->debug)
    app->diagMap(nidx);

  this->success = false;
  wl = this->getWatchList(conf);
  if (!wl->success) {
    this->problem = wl->problem;
    return false;
    }
  this->conf    = conf;
  this->nhosts  = nfiles = (int) wl->watch.size();
  app->namelen  = wl->namelen;

  this->notify = inotify_init();
  if (this->notify < 0)
    return this->setError("notify_init");
  this->hosts  = inst = new vhost[nfiles];
  this->hlist  = new vhost[nfiles];
  k = vhost::OUTPUT_FIELDS + app->needs.size();
  for(i=0,iter=wl->watch.begin();iter!=wl->watch.end();++iter,i++) {
    for(j=0;j<k;j++)
      inst[i].fields.push_back("");
    inst[i].fields[vhost::FILE_FIELD]   = iter->second;
    inst[i].fields[inst[i].nidx = nidx] = iter->first;
    inst[i].wd  = inotify_add_watch(this->notify,inst[i].fields[1].c_str(),IN_MODIFY);
    this->watch[inst[i].wd] = i;
    if (!inst[i].setDirty(options.init))
      return false;
    }
  this->success = true;
  return true;
}
/**
 * Parse apache log format string to build a vector of identitifers for the white
 * space separated log record components. The output is used to identifer parameter
 * position and type - since this application only recognizes a few types, modifiers
 * and literals such as quotes can be ignored.
 */
void Monitor::parseApacheLogFormat(topvhost *app)
{
  char   ch;
  bool   quoted = false;
  bool   quotearg = false;
  string suffix="cinpPt";
  string token;
  string::iterator format;

  if (app->logfmt.length() == 0)
    app->logfmt = FMT_DEFAULT;
  format = app->logfmt.begin();
  do {
    ch = format==app->logfmt.end()? 0 : *format++;
    if (ch=='\\' && format!=app->logfmt.end()) ch=*format++;
    switch(ch) {
      case '"':
        quoted = !quoted;
        break;
      case '%':
        quotearg = quoted;
        token=ch;
        break;
      case 0: case ' ':
        if (!quoted) {
          if (token[0]=='%') {
            int len = token.length();
            if (len>2 && suffix.find(token[len-1])!=string::npos)
              app->gets.push_back(token.substr(len-1)+token.substr(1,len-2));
            else app->gets.push_back(token.substr(len-1));
            }
          else app->gets.push_back("");
          token="";
          }
        break;
      default:
        token += ch;
      }
    } while(ch !=0);
}
/**
 * Terminate the monitor
 */
void Monitor::quit(void)
{
  close(this->notify);
}
/**
 * Force a display update
 */
void Monitor::refresh(void)
{
  int i;

  for(i=0;i<this->nhosts;i++)
    this->hosts[i].dirty = true;
}
/**
 * Set error message
 */
bool Monitor::setError(char *tag)
{
  char msg[256];

  strcpy(msg,tag),
  strerror_r(errno, msg+strlen(tag), 255-strlen(tag));
  this->problem = msg;
  return false;
}
/**
 * Update the list - update the last date time modified and set the dirty bit
 * for inotify events received
 */
bool Monitor::update(topvhost *app)
{
  char buf[BUF_LEN];
  struct inotify_event *event;
  fd_set read_fd;
  struct timeval tv;
  int len, i, idx, rc;

  FD_ZERO(&read_fd);
  FD_SET(this->notify,&read_fd);
  tv.tv_sec = tv.tv_usec = 0;
  rc = select(this->notify+1, &read_fd, NULL, NULL, &tv);
  if (rc<=0) {
    if (rc<0 && errno != EINTR)
      return this->setError("select:");
    return true;
    }
  len = read (this->notify, buf, BUF_LEN);
  if (len<0 && errno!=EINTR)
    return this->setError("read:");
  for(i=0;i<len;i+=EVENT_SIZE + event->len) {
    event = (struct inotify_event *) &buf[i];
    idx = this->watch[event->wd];
    if ((event->mask & IN_MODIFY) != 0 && !this->hosts[idx].setDirty())
      this->setError("stat:");
    }
  return true;
}
